import argparse
import copy
import html
import logging
import os
import re
import sys
from collections.abc import MutableMapping, MutableSequence
from io import StringIO, TextIOWrapper
from typing import IO, Any, Literal, Optional, Union, cast
from urllib.parse import urldefrag

from mistune import create_markdown
from mistune.markdown import Markdown
from mistune.renderers.html import HTMLRenderer

from .exceptions import SchemaSaladException, ValidationException
from .schema import avro_field_name, extend_and_specialize, get_metaschema
from .utils import add_dictlist, aslist
from .validate import avro_type_name

PluginName = Literal[
    "url",
    "strikethrough",
    "footnotes",
    "table",
    "task_lists",
    "def_list",
    "abbr",
]

_logger = logging.getLogger("salad")


def escape_html(s: str) -> str:
    """Escape HTML but otherwise preserve single quotes."""
    return html.escape(html.unescape(s)).replace("&#x27;", "'")


def vocab_type_name(url: str) -> str:
    """Remove the avro namespace, if any."""
    return avro_type_name(url).split(".")[-1]


def has_types(items: Any) -> list[str]:
    """Retrieve all the types of a record."""
    r: list[str] = []
    if isinstance(items, MutableMapping):
        if items["type"] == "https://w3id.org/cwl/salad#record":
            return [items["name"]]
        for n in ("type", "items", "values"):
            if n in items:
                r.extend(has_types(items[n]))
        return r
    if isinstance(items, MutableSequence):
        for i in items:
            r.extend(has_types(i))
        return r
    if isinstance(items, str):
        return [items]
    return []


def linkto(item: str) -> str:
    frg = urldefrag(item)[1]
    return f"[{frg}](#{to_id(frg)})"


class MyRenderer(HTMLRenderer):
    """Custom renderer with different representations of selected HTML tags."""

    def heading(self, text: str, level: int, **attrs: Any) -> str:
        """Override HTML heading creation with text IDs."""
        return """<h{} id="{}" class="section">{} <a href="#{}">&sect;</a></h{}>""".format(
            level, to_id(text), text, to_id(text), level
        )

    def text(self, text: str) -> str:
        """Don't escape quotation marks."""
        # avoid convert of & if already escaped
        text = html.unescape(text)
        # html.escape does both single/double quotes ('/")
        # mistune.util.escape does only double quotes
        return html.escape(text, quote=self._escape)

    def inline_html(self, html: str) -> str:
        """Don't escape characters in predefined HTML within paragraph tags."""
        return html + "\n"

    def block_html(self, html: str) -> str:
        """Don't escape characters nor wrap predefined HTML within paragraph tags."""
        return html + "\n"

    def block_code(self, code: str, info: Optional[str] = None) -> str:
        """Don't escape quotation marks."""
        text = "<pre><code"
        if info is not None:
            info = info.strip()
        if info:
            lang = info.split(None, 1)[0]
            lang = escape_html(lang)
            text += ' class="language-' + lang + '"'
        return text + ">" + html.escape(code, quote=self._escape) + "</code></pre>\n"


def patch_fenced_code(original_markdown_text: str, modified_markdown_text: str) -> str:
    """Reverts fenced code fragments found in the modified contents back to their original definition."""
    # Pattern inspired from 'mistune.block_parser.BlockParser.FENCED_CODE'.
    # However, instead of the initial ' {0,3}' part to match any indented fenced-code,
    # use any quantity of spaces, as long as they match at the end as well (using '\1').
    # Because of nested fenced-code in lists, it can be more indented than "normal".
    fenced_code_pattern = re.compile(
        r"( *)(`{3,}|~{3,})([^`\n]*)\n(?:|([\s\S]*?)\n)(?:\1\2[~`]* *\n+|$)"
    )
    matches_original = list(re.finditer(fenced_code_pattern, original_markdown_text))
    matches_modified = list(re.finditer(fenced_code_pattern, modified_markdown_text))
    if len(matches_original) != len(matches_modified):
        raise ValueError("Cannot patch fenced code definitions with inconsistent matches.")
    result = ""
    begin = 0
    for original, modified in zip(matches_original, matches_modified):
        ori_s, ori_e = original.start(), original.end()
        mod_s, mod_e = modified.start(), modified.end()
        result += modified_markdown_text[begin:mod_s]  # add text in between matches
        result += original_markdown_text[ori_s:ori_e]  # revert the fenced code
        begin = mod_e  # skip over the modified fenced code for next match
    result += modified_markdown_text[begin:]  # left over text after last match
    return result


def to_id(text: str) -> str:
    textid = text
    if text[0] in ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9"):
        try:
            textid = text[text.index(" ") + 1 :]
        except ValueError:
            pass
    return textid.replace(" ", "_")


class ToC:
    def __init__(self) -> None:
        self.first_toc_entry = True
        self.numbering = [0]
        self.toc = ""
        self.start_numbering = True

    def add_entry(self, thisdepth: int, title: str) -> str:
        """Add an entry to the table of contents."""
        depth = len(self.numbering)
        if thisdepth < depth:
            self.toc += "</ol>"
            for _ in range(0, depth - thisdepth):
                self.numbering.pop()
                self.toc += "</li></ol>"
            self.numbering[-1] += 1
        elif thisdepth == depth:
            if not self.first_toc_entry:
                self.toc += "</ol>"
            else:
                self.first_toc_entry = False
            self.numbering[-1] += 1
        elif thisdepth > depth:
            self.numbering.append(1)

        num = (
            "{}.{}".format(self.numbering[0], ".".join([str(n) for n in self.numbering[1:]]))
            if self.start_numbering
            else ""
        )
        self.toc += f"""<li><a href="#{to_id(title)}">{num} {title}</a><ol>\n"""
        return num

    def contents(self, idn: str) -> str:
        toc = """<h1 id="{}">Table of contents</h1>
               <nav class="tocnav"><ol>{}""".format(
            idn, self.toc
        )
        toc += "</ol>"
        for _ in range(0, len(self.numbering)):
            toc += "</li></ol>"
        toc += """</nav>"""
        return toc


basicTypes = (
    "https://w3id.org/cwl/salad#null",
    "http://www.w3.org/2001/XMLSchema#boolean",
    "http://www.w3.org/2001/XMLSchema#int",
    "http://www.w3.org/2001/XMLSchema#long",
    "http://www.w3.org/2001/XMLSchema#float",
    "http://www.w3.org/2001/XMLSchema#double",
    "http://www.w3.org/2001/XMLSchema#string",
    "https://w3id.org/cwl/salad#record",
    "https://w3id.org/cwl/salad#enum",
    "https://w3id.org/cwl/salad#array",
)


def number_headings(toc: ToC, maindoc: str) -> str:
    mdlines = []
    skip = False
    for line in maindoc.splitlines():
        if line.strip() == "# Introduction":
            toc.start_numbering = True
            toc.numbering.clear()
            toc.numbering.append(0)

        if "```" in line:
            skip = not skip

        if not skip:
            m = re.match(r"^(#+) (.*)", line)
            if m is not None:
                group1 = m.group(1)
                assert group1 is not None  # nosec
                group2 = m.group(2)
                assert group2 is not None  # nosec
                num = toc.add_entry(len(group1), group2)
                line = f"{group1} {num} {group2}"
            line = re.sub(r"^(https?://\S+)", r"[\1](\1)", line)
        mdlines.append(line)

    maindoc = "\n".join(mdlines)
    return maindoc


def fix_doc(doc: Union[list[str], str]) -> str:
    """Concatenate doc strings, replacing email addresses with mailto links."""
    docstr = "".join(doc) if isinstance(doc, MutableSequence) else doc
    return "\n".join(
        [re.sub(r"<([^>@]+@[^>]+)>", r"[\1](mailto:\1)", d) for d in docstr.splitlines()]
    )


class RenderType:
    def __init__(
        self,
        toc: ToC,
        j: list[dict[str, Any]],
        renderlist: list[str],
        redirects: dict[str, str],
        primitiveType: str,
    ) -> None:
        self.typedoc = StringIO()
        self.toc = toc
        self.subs: dict[str, str] = {}
        self.docParent: dict[str, list[str]] = {}
        self.docAfter: dict[str, list[str]] = {}
        self.rendered: set[str] = set()
        self.redirects = redirects
        self.title: Optional[str] = None
        self.primitiveType = primitiveType

        for t in j:
            if "extends" in t:
                for e in aslist(t["extends"]):
                    add_dictlist(self.subs, e, t["name"])
                    # if "docParent" not in t and "docAfter" not in t:
                    #    add_dictlist(self.docParent, e, t["name"])

            if t.get("docParent"):
                add_dictlist(self.docParent, t["docParent"], t["name"])

            if t.get("docChild"):
                for c in aslist(t["docChild"]):
                    add_dictlist(self.docParent, t["name"], c)

            if t.get("docAfter"):
                add_dictlist(self.docAfter, t["docAfter"], t["name"])

        metaschema_loader = get_metaschema()[2]
        alltypes = extend_and_specialize(j, metaschema_loader)

        self.typemap: dict[str, dict[str, str]] = {}
        self.uses: dict[str, list[tuple[str, str]]] = {}
        self.record_refs: dict[str, list[str]] = {}
        for entry in alltypes:
            self.typemap[entry["name"]] = entry
            try:
                if entry["type"] == "record":
                    self.record_refs[entry["name"]] = []
                    fields: Union[str, list[dict[str, str]]] = entry.get("fields", [])
                    if isinstance(fields, str):
                        raise KeyError("record fields must be a list of mappings")
                    else:
                        for f in fields:
                            p = has_types(f)
                            for tp in p:
                                if tp not in self.uses:
                                    self.uses[tp] = []
                                if (entry["name"], f["name"]) not in self.uses[tp]:
                                    _, frg1 = urldefrag(t["name"])
                                    _, frg2 = urldefrag(f["name"])
                                    self.uses[tp].append((frg1, frg2))
                                if (
                                    tp not in basicTypes
                                    and tp not in self.record_refs[entry["name"]]
                                ):
                                    self.record_refs[entry["name"]].append(tp)
            except KeyError:
                _logger.error("Did not find 'type' in %s", t)
                _logger.error("record refs is %s", self.record_refs)
                raise

        for entry in alltypes:
            if entry["name"] in renderlist or (
                (not renderlist)
                and ("extends" not in entry)
                and ("docParent" not in entry)
                and ("docAfter" not in entry)
            ):
                self.render_type(entry, 1)

    def typefmt(
        self,
        tp: Any,
        redirects: dict[str, str],
        nbsp: bool = False,
        jsonldPredicate: Optional[Union[dict[str, str], str]] = None,
    ) -> str:
        if isinstance(tp, MutableSequence):
            if nbsp and len(tp) <= 3:
                return "&nbsp;|&nbsp;".join(
                    [self.typefmt(n, redirects, jsonldPredicate=jsonldPredicate) for n in tp]
                )
            return " | ".join(
                [self.typefmt(n, redirects, jsonldPredicate=jsonldPredicate) for n in tp]
            )
        if isinstance(tp, MutableMapping):
            if tp["type"] == "https://w3id.org/cwl/salad#array":
                ar = "array&lt;{}&gt;".format(self.typefmt(tp["items"], redirects, nbsp=True))
                if isinstance(jsonldPredicate, dict) and "mapSubject" in jsonldPredicate:
                    if "mapPredicate" in jsonldPredicate:
                        ar += " | "
                        if len(ar) > 40:
                            ar += "<br>"

                        ar += (
                            "<a href='#map'>map</a>&lt;<code>{}</code>"
                            ",&nbsp;<code>{}</code> | {}&gt".format(
                                jsonldPredicate["mapSubject"],
                                jsonldPredicate["mapPredicate"],
                                self.typefmt(tp["items"], redirects),
                            )
                        )
                    else:
                        ar += " | "
                        if len(ar) > 40:
                            ar += "<br>"
                        ar += "<a href='#map'>map</a>&lt;<code>{}</code>,&nbsp;{}&gt".format(
                            jsonldPredicate["mapSubject"],
                            self.typefmt(tp["items"], redirects),
                        )
                return ar
            if tp["type"] in (
                "https://w3id.org/cwl/salad#record",
                "https://w3id.org/cwl/salad#enum",
            ):
                frg = vocab_type_name(tp["name"])
                if tp["name"] in redirects:
                    return """<a href="{}">{}</a>""".format(redirects[tp["name"]], frg)
                if tp["name"] in self.typemap:
                    return f"""<a href="#{to_id(frg)}">{frg}</a>"""
                if tp["type"] == "https://w3id.org/cwl/salad#enum" and len(tp["symbols"]) == 1:
                    return "constant value <code>{}</code>".format(
                        avro_field_name(tp["symbols"][0])
                    )
                return frg
            if isinstance(tp["type"], MutableMapping):
                return self.typefmt(tp["type"], redirects)
        else:
            if str(tp) in redirects:
                return f"""<a href="{redirects[tp]}">{redirects[tp]}</a>"""  # noqa: B907
            if str(tp) in basicTypes:
                return """<a href="{}">{}</a>""".format(
                    self.primitiveType, vocab_type_name(str(tp))
                )
            frg2 = urldefrag(tp)[1]
            if frg2 != "":
                tp = frg2
            return f"""<a href="#{to_id(tp)}">{tp}</a>"""
        raise SchemaSaladException("We should not be here!")

    def render_type(self, f: dict[str, Any], depth: int) -> None:
        """Render a type declaration."""
        if f["name"] in self.rendered or f["name"] in self.redirects:
            return
        self.rendered.add(f["name"])

        if f.get("abstract"):
            return

        if "doc" not in f:
            f["doc"] = ""

        f["type"] = copy.deepcopy(f)
        f["doc"] = ""
        f = f["type"]

        if "doc" not in f:
            f["doc"] = ""

        def extendsfrom(item: dict[str, Any], ex: list[dict[str, Any]]) -> None:
            if "extends" in item:
                for e in aslist(item["extends"]):
                    ex.insert(0, self.typemap[e])
                    extendsfrom(self.typemap[e], ex)

        ex = [f]
        extendsfrom(f, ex)

        enumDesc = {}
        if f["type"] == "enum" and isinstance(f["doc"], MutableSequence):
            for e in ex:
                for i in e["doc"]:
                    idx = i.find(":")
                    if idx > -1:
                        enumDesc[i[:idx]] = i[idx + 1 :]
                e["doc"] = [i for i in e["doc"] if i.find(":") == -1 or i.find(" ") < i.find(":")]

        f["doc"] = fix_doc(f["doc"])

        if f["type"] == "record":
            for field in f.get("fields", []):
                if "doc" not in field:
                    field["doc"] = ""

        if f["type"] != "documentation":
            lines = []
            for line in f["doc"].splitlines():
                if len(line) > 0 and line[0] == "#":
                    line = ("#" * depth) + line
                lines.append(line)
            f["doc"] = "\n".join(lines)

            frg = urldefrag(f["name"])[1]
            num = self.toc.add_entry(depth, frg)
            doc = "{} {} {}\n".format(("#" * depth), num, frg)
        else:
            doc = ""

        # Save the first line of the first type definition for the
        # HTML <title> tag
        if self.title is None and f["doc"]:
            self.title = f["doc"].partition("\n")[0]
            if self.title.startswith("# "):
                self.title = self.title[2:]

        if f["type"] == "documentation":
            f["doc"] = number_headings(self.toc, f["doc"])

        doc = doc + "\n\n" + f["doc"]
        plugins: list[PluginName] = [
            "strikethrough",
            "footnotes",
            "table",
            "url",
        ]
        # if escape active, wraps literal HTML into '<p> {HTML} </p>'
        # we must pass it to both since 'MyRenderer' is predefined
        escape = False
        markdown2html: Markdown = create_markdown(
            renderer=MyRenderer(escape=escape),
            plugins=plugins,
            escape=escape,
        )
        doc = cast(str, markdown2html(doc))

        if f["type"] == "record":
            doc += "<h3>Fields</h3>"
            doc += """
<div class="responsive-table">
<div class="row responsive-table-header">
<div class="col-xs-3 col-lg-2">field</div>
<div class="col-xs-2 col-lg-1">required</div>
<div class="col-xs-7 col-lg-3">type</div>
<div class="col-xs-12 col-lg-6 description-header">description</div>
</div>"""
            required = []
            optional = []
            for i in f.get("fields", []):
                tp = i["type"]
                if isinstance(tp, MutableSequence) and tp[0] == "https://w3id.org/cwl/salad#null":
                    opt = False
                    tp = tp[1:]
                else:
                    opt = True

                desc = i["doc"]

                rfrg = avro_field_name(i["name"])
                tr = """
<div class="row responsive-table-row">
<div class="col-xs-3 col-lg-2"><code>{}</code></div>
<div class="col-xs-2 col-lg-1">{}</div>
<div class="col-xs-7 col-lg-3">{}</div>
<div class="col-xs-12 col-lg-6 description-col">{}</div>
</div>""".format(
                    rfrg,
                    "required" if opt else "optional",
                    self.typefmt(tp, self.redirects, jsonldPredicate=i.get("jsonldPredicate")),
                    markdown2html(desc),
                )
                if opt:
                    required.append(tr)
                else:
                    optional.append(tr)
            for i in required + optional:
                doc += i
            doc += """</div>"""
        elif f["type"] == "enum":
            doc += "<h3>Symbols</h3>"
            doc += """<table class="table table-striped">"""
            doc += "<tr><th>symbol</th><th>description</th></tr>"
            for e in ex:
                for i in e.get("symbols", []):
                    doc += "<tr>"
                    efrg = avro_field_name(i)
                    doc += "<td><code>{}</code></td><td>{}</td>".format(
                        efrg, enumDesc.get(efrg, "")
                    )
                    doc += "</tr>"
            doc += """</table>"""
        f["doc"] = doc

        self.typedoc.write(f["doc"])

        subs = self.docParent.get(f["name"], []) + self.record_refs.get(f["name"], [])
        if len(subs) == 1:
            self.render_type(self.typemap[subs[0]], depth)
        else:
            for s in subs:
                self.render_type(self.typemap[s], depth + 1)

        for s in self.docAfter.get(f["name"], []):
            self.render_type(self.typemap[s], depth)


def avrold_doc(
    j: list[dict[str, Any]],
    outdoc: IO[Any],
    renderlist: list[str],
    redirects: dict[str, str],
    brand: str,
    brandlink: str,
    primtype: str,
    brandstyle: Optional[str] = None,
    brandinverse: Optional[bool] = False,
) -> None:
    toc = ToC()
    toc.start_numbering = False

    rt = RenderType(toc, j, renderlist, redirects, primtype)
    content = rt.typedoc.getvalue()

    if brandstyle is None:
        bootstrap_url = "https://cdn.jsdelivr.net/npm/bootstrap@5.0.2/dist/css/bootstrap.min.css"
        bootstrap_integrity = (
            "sha384-EVSTQN3/azprG1Anm3QDgpJLIm9Nao0Yz1ztcQTwFspd3yD65VohhpuuCOmLASjC"
        )
        brandstyle_template = '<link rel="stylesheet" href={} integrity={} crossorigin="anonymous">'
        brandstyle = brandstyle_template.format(bootstrap_url, bootstrap_integrity)

    picturefill_url = "https://cdn.rawgit.com/scottjehl/picturefill/3.0.2/dist/picturefill.min.js"
    picturefill_integrity = (
        "sha384-ZJsVW8YHHxQHJ+SJDncpN90d0EfAhPP+yA94n+EhSRzhcxfo84yMnNk+v37RGlWR"
    )
    outdoc.write(
        """
    <!DOCTYPE html>
    <html>
    <head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    {}
    <script>
    // Picture element HTML5 shiv
    document.createElement( "picture" );
    </script>
    <script src="{}"
        integrity="{}"
        crossorigin="anonymous" async></script>
    """.format(
            brandstyle, picturefill_url, picturefill_integrity
        )
    )

    outdoc.write(f"<title>{rt.title}</title>")

    outdoc.write(
        """
    <style>
    :target {
      padding-top: 61px;
      margin-top: -61px;
    }
    .tocnav ol {
      list-style: none
    }
    pre {
      margin: 0 2em 10px 2em;
      padding: 9.5px;
      line-height: 1.42857143;
      color: #333;
      word-break: break-all;
      word-wrap: break-word;
      background-color: #f5f5f5;
      border: 1px solid #ccc;
      border-radius: 4px;
    }
    pre code {
      padding: 0;
      font-size: inherit;
      color: inherit;
      white-space: pre-wrap;
      background-color: transparent;
      border-radius: 0;
    }
    code {
      background-color: #f9f2f4;
      border-radius: 4px;
      padding: 2px 4px;
      color: #c7254e;
    }
    blockquote {
      padding: 10px 20px 1px 20px;
      border-left: 5px solid #eee;
    }
    a {
      text-decoration: none;
    }
    a code {
      color: #c7254e;
    }
    .section a {
      visibility: hidden;
    }
    .section:hover a {
      visibility: visible;
      color: rgb(201, 201, 201);
    }
    .responsive-table-header {
      text-align: left;
      padding: 8px;
      vertical-align: top;
      font-weight: bold;
      border-top-color: rgb(221, 221, 221);
      border-top-style: solid;
      border-top-width: 1px;
      background-color: #f9f9f9
    }
    .responsive-table > .responsive-table-row {
      text-align: left;
      padding: 8px;
      vertical-align: top;
      border-top-color: rgb(221, 221, 221);
      border-top-style: solid;
      border-top-width: 1px;
    }
    @media (min-width: 0px), print {
      .description-header {
        display: none;
      }
      .description-col {
        margin-top: 1em;
        margin-left: 1.5em;
      }
    }
    @media (min-width: 1170px) {
      .description-header {
        display: inline;
      }
      .description-col {
        margin-top: 0px;
        margin-left: 0px;
      }
    }
    .responsive-table-row:nth-of-type(odd) {
       background-color: #f9f9f9
    }
    </style>
    </head>
    <body>
    """
    )

    navbar_extraclass = "navbar-inverse" if brandinverse else ""
    outdoc.write(
        """
      <nav class="navbar sticky-top navbar-expand-lg navbar-light bg-light {}">
        <div class="container">
          <a class="navbar-brand" href="{}">{}</a>
    """.format(
            navbar_extraclass, brandlink, brand
        )
    )

    if "<!--ToC-->" in content:
        content = content.replace("<!--ToC-->", toc.contents("toc"))
        outdoc.write(
            """
              <ul class="navbar-nav me-auto">
                <li class="nav-item"><a class="nav-link" href="#toc">Table of contents</a></li>
              </ul>
        """
        )

    outdoc.write(
        """
        </div>
      </nav>
    """
    )

    outdoc.write(
        """
    <div class="container mt-4">
    """
    )

    outdoc.write(
        """
    <div class="row">
    """
    )

    outdoc.write(
        """
    <div class="col-md-12" role="main" id="main">"""
    )

    outdoc.write(content)

    outdoc.write("""</div>""")

    outdoc.write(
        """
    </div>
    </div>
    </body>
    </html>"""
    )


def arg_parser() -> argparse.ArgumentParser:
    """Build the argument parser."""
    parser = argparse.ArgumentParser()
    parser.add_argument("schema")
    parser.add_argument("--only", action="append")
    parser.add_argument("--redirect", action="append")
    parser.add_argument("--brand")
    parser.add_argument("--brandlink")
    parser.add_argument("--brandstyle")
    parser.add_argument("--brandinverse", default=False, action="store_true")
    parser.add_argument("--primtype", default="#PrimitiveType")
    parser.add_argument("--debug", action="store_true")
    return parser


def main() -> None:
    """Shortcut entrypoint."""
    args = arg_parser().parse_args()
    if args.debug:
        _logger.setLevel(logging.DEBUG)
    makedoc(
        sys.stdout,
        args.schema,
        args.redirect,
        args.only,
        args.brand,
        args.brandlink,
        args.primtype,
        args.brandstyle,
        args.brandinverse,
    )


def makedoc(
    stdout: IO[Any],
    schema: str,
    redirects: Optional[list[str]] = None,
    only: Optional[list[str]] = None,
    brand: Optional[str] = None,
    brandlink: Optional[str] = None,
    primtype: Optional[str] = None,
    brandstyle: Optional[str] = None,
    brandinverse: Optional[bool] = False,
) -> None:
    """Emit HTML representation of a given schema."""
    s: list[dict[str, Any]] = []
    with open(schema, encoding="utf-8") as f:
        if schema.endswith("md"):
            s.append(
                {
                    "name": os.path.splitext(os.path.basename(schema))[0],
                    "type": "documentation",
                    "doc": f.read(),
                }
            )
        else:
            uri = "file://" + os.path.abspath(schema)
            metaschema_loader = get_metaschema()[2]
            j = metaschema_loader.resolve_ref(uri, "")[0]
            if isinstance(j, MutableSequence):
                s.extend(j)
            elif isinstance(j, MutableMapping):
                s.append(j)
            else:
                raise ValidationException("Schema must resolve to a list or a dict")
    redirect = {}
    for r in redirects or []:
        redirect[r.split("=")[0]] = r.split("=")[1]
    renderlist = only or []
    if hasattr(stdout, "buffer") and getattr(stdout, "encoding", None) != "UTF-8":
        wrapped_stdout: IO[Any] = TextIOWrapper(stdout.buffer, encoding="utf-8")
    else:
        wrapped_stdout = stdout
    avrold_doc(
        s,
        wrapped_stdout,
        renderlist,
        redirect,
        brand or "",
        brandlink or "",
        primtype or "",
        brandstyle=brandstyle,
        brandinverse=brandinverse,
    )


if __name__ == "__main__":
    main()



import re
from collections.abc import Generator as Generator
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    List,
    Match,
    MutableMapping,
    Optional,
    Pattern,
    Type,
    TypeVar,
    Union,
)

from typing_extensions import Self

class BlockState:
    src: str
    tokens: List[Dict[str, Any]]
    cursor: int
    cursor_max: int
    list_tight: bool
    parent: Any
    env: MutableMapping[str, Any]
    def __init__(self, parent: Optional[Any] = None) -> None: ...
    def child_state(self, src: str) -> BlockState: ...
    def process(self, src: str) -> None: ...
    def find_line_end(self) -> int: ...
    def get_text(self, end_pos: int) -> str: ...
    def last_token(self) -> Any: ...
    def prepend_token(self, token: Dict[str, Any]) -> None: ...
    def append_token(self, token: Dict[str, Any]) -> None: ...
    def add_paragraph(self, text: str) -> None: ...
    def append_paragraph(self) -> Optional[int]: ...
    def depth(self) -> int: ...

class InlineState:
    env: MutableMapping[str, Any]
    src: str
    tokens: List[Dict[str, Any]]
    in_image: bool
    in_link: bool
    in_emphasis: bool
    in_strong: bool
    def __init__(self, env: MutableMapping[str, Any]) -> None: ...
    def prepend_token(self, token: Dict[str, Any]) -> None: ...
    def append_token(self, token: Dict[str, Any]) -> None: ...
    def copy(self) -> InlineState: ...

ST = TypeVar("ST", InlineState, BlockState)

class Parser(Generic[ST]):
    sc_flag: re._FlagsType
    state_cls: Type[ST]
    SPECIFICATION: ClassVar[Dict[str, str]]
    DEFAULT_RULES: ClassVar[Iterable[str]]
    specification: Dict[str, str]
    rules: Iterable[str]
    def __init__(self) -> None: ...
    def compile_sc(self, rules: Optional[List[str]] = None) -> Pattern[str]: ...
    def register(
        self,
        name: str,
        pattern: Union[str, None],
        func: Callable[[Self, Match[str], ST], Optional[int]],
        before: Optional[str] = None,
    ) -> None: ...
    def register_rule(self, name: str, pattern: str, func: Any) -> None: ...
    @staticmethod
    def insert_rule(rules: List[str], name: str, before: Optional[str] = None) -> None: ...
    def parse_method(self, m: Match[str], state: ST) -> Optional[int]: ...

class BaseRenderer:
    NAME: ClassVar[str]
    def __init__(self) -> None: ...
    def register(self, name: str, method: Callable[..., str]) -> None: ...
    def render_token(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def iter_tokens(self, tokens: Iterable[Dict[str, Any]], state: BlockState) -> Iterable[str]: ...
    def render_tokens(self, tokens: Iterable[Dict[str, Any]], state: BlockState) -> str: ...
    def __call__(self, tokens: Iterable[Dict[str, Any]], state: BlockState) -> str: ...



"""Python code generator for a given schema salad definition."""

import textwrap
from collections.abc import MutableMapping, MutableSequence
from io import StringIO
from typing import IO, Any, Optional, Union

try:
    import black
except ModuleNotFoundError:
    black = None  # type: ignore[assignment]

from . import schema
from .codegen_base import CodeGenBase, LazyInitDef, TypeDef
from .exceptions import SchemaException
from .schema import shortname
from .utils import files

_string_type_def = TypeDef("strtype", "_PrimitiveLoader(str)")
_int_type_def = TypeDef("inttype", "_PrimitiveLoader(int)")
_float_type_def = TypeDef("floattype", "_PrimitiveLoader(float)")
_bool_type_def = TypeDef("booltype", "_PrimitiveLoader(bool)")
_null_type_def = TypeDef("None_type", "_PrimitiveLoader(type(None))")
_any_type_def = TypeDef("Any_type", "_AnyLoader()")

prims = {
    "http://www.w3.org/2001/XMLSchema#string": _string_type_def,
    "http://www.w3.org/2001/XMLSchema#int": _int_type_def,
    "http://www.w3.org/2001/XMLSchema#long": _int_type_def,
    "http://www.w3.org/2001/XMLSchema#float": _float_type_def,
    "http://www.w3.org/2001/XMLSchema#double": _float_type_def,
    "http://www.w3.org/2001/XMLSchema#boolean": _bool_type_def,
    "https://w3id.org/cwl/salad#null": _null_type_def,
    "https://w3id.org/cwl/salad#Any": _any_type_def,
    "string": _string_type_def,
    "int": _int_type_def,
    "long": _int_type_def,
    "float": _float_type_def,
    "double": _float_type_def,
    "boolean": _bool_type_def,
    "null": _null_type_def,
    "Any": _any_type_def,
}


def fmt(text: str, indent: int) -> str:
    """
    Use black to format this snippet.

    :param indent the indent level for the current context
    """
    if not black:
        raise Exception(
            "Must install 'black' to use the Python code generator. "
            "If installing schema-salad via pip, try `pip install schema-salad[pycodegen]`."
        )
    return textwrap.indent(
        black.format_str(
            text,
            mode=black.mode.Mode(
                target_versions={black.mode.TargetVersion.PY36}, line_length=88 - indent
            ),
        ),
        " " * indent,
    )


class PythonCodeGen(CodeGenBase):
    """Generation of Python code for a given Schema Salad definition."""

    def __init__(
        self,
        out: IO[str],
        copyright: Optional[str],
        parser_info: str,
        salad_version: str,
    ) -> None:
        super().__init__()
        self.out = out
        self.current_class_is_abstract = False
        self.serializer = StringIO()
        self.idfield = ""
        self.copyright = copyright
        self.parser_info = parser_info
        self.salad_version = salad_version

    @staticmethod
    def safe_name(name: str) -> str:
        """Generate a safe version of the given name."""
        avn = schema.avro_field_name(name)
        if avn.startswith("anon."):
            avn = avn[5:]
        elif avn[0].isdigit():
            avn = f"_{avn}"
        elif avn in ("class", "in", "type"):
            # reserved words
            avn = f"{avn}_"
        return avn.replace(".", "_")

    def prologue(self) -> None:
        """Trigger to generate the prolouge code."""
        self.out.write(
            """#
# This file was autogenerated using schema-salad-tool --codegen=python
# The code itself is released under the Apache 2.0 license and the help text is
# subject to the license of the original schema.
"""
        )
        if self.copyright:
            self.out.write(
                """
#
# The original schema is {copyright}.
""".format(
                    copyright=self.copyright
                )
            )

        python_codegen_support = (
            files("schema_salad").joinpath("python_codegen_support.py").read_text("UTF-8")
        )
        self.out.write(python_codegen_support[python_codegen_support.find("\n") + 1 :])
        self.out.write("\n\n")

        self.out.write(
            f"""def parser_info() -> str:
    return "{self.parser_info}"


"""  # noqa: B907
        )

        for primitive in prims.values():
            self.declare_type(primitive)

    def begin_class(
        self,  # pylint: disable=too-many-arguments
        classname: str,
        extends: MutableSequence[str],
        doc: str,
        abstract: bool,
        field_names: MutableSequence[str],
        idfield: str,
        optional_fields: set[str],
    ) -> None:
        classname = self.safe_name(classname)

        if extends:
            ext = ", ".join(self.safe_name(e) for e in extends)
        else:
            ext = "Saveable"

        self.out.write(fmt(f"class {classname}({ext}):\n    pass", 0)[:-9])
        # make a valid class for Black, but then trim off the "pass"

        if doc:
            self.out.write(fmt(f'"""\n{doc}\n"""\n', 4) + "\n")

        self.serializer = StringIO()

        self.current_class_is_abstract = abstract
        if self.current_class_is_abstract:
            self.out.write("    pass\n\n\n")
            return

        idfield_safe_name = self.safe_name(idfield) if idfield != "" else None
        if idfield_safe_name is not None:
            self.out.write(f"    {idfield_safe_name}: str\n\n")

        required_field_names = [f for f in field_names if f not in optional_fields]
        optional_field_names = [f for f in field_names if f in optional_fields]

        safe_inits: list[str] = ["        self,"]
        safe_inits.extend(
            [f"        {self.safe_name(f)}: Any," for f in required_field_names if f != "class"]
        )
        safe_inits.extend(
            [
                f"        {self.safe_name(f)}: Optional[Any] = None,"
                for f in optional_field_names
                if f != "class"
            ]
        )
        self.out.write(
            "    def __init__(\n"
            + "\n".join(safe_inits)
            + "\n        extension_fields: Optional[dict[str, Any]] = None,"
            + "\n        loadingOptions: Optional[LoadingOptions] = None,"
            + "\n    ) -> None:\n"
            + """        if extension_fields:
            self.extension_fields = extension_fields
        else:
            self.extension_fields = CommentedMap()
        if loadingOptions:
            self.loadingOptions = loadingOptions
        else:
            self.loadingOptions = LoadingOptions()
"""
        )
        field_inits = ""
        for name in field_names:
            if name == "class":
                field_inits += """        self.class_ = "{}"
""".format(
                    classname
                )
            elif name == idfield_safe_name:
                field_inits += """        self.{0} = {0} if {0} is not None else "_:" + str(_uuid__.uuid4())
""".format(
                    self.safe_name(name)
                )
            else:
                field_inits += """        self.{0} = {0}
""".format(
                    self.safe_name(name)
                )
        field_eqs = []
        field_hashes = []
        for name in field_names:
            field_eqs.append("self.{0} == other.{0}".format(self.safe_name(name)))
            field_hashes.append(f"self.{self.safe_name(name)}")
        field_eq = " and\n                    ".join(field_eqs)
        field_hash = ",\n            ".join(field_hashes)
        self.out.write(field_inits)
        self.out.write(
            "\n"
            + fmt(
                f"""def __eq__(
    self,
    other: Any
) -> bool:
    if isinstance(other, {classname}):
        return bool({field_eq})
    return False
""",
                4,
            )
        )
        self.out.write(
            "\n"
            + fmt(
                f"""def __hash__(self) -> int:
    return hash((
        {field_hash}
    ))
""",
                4,
            )
        )

        self.out.write(
            f"""
    @classmethod
    def fromDoc(
        cls,
        doc: Any,
        baseuri: str,
        loadingOptions: LoadingOptions,
        docRoot: Optional[str] = None
    ) -> "{classname}":
        _doc = copy.copy(doc)

        if hasattr(doc, "lc"):
            _doc.lc.data = doc.lc.data
            _doc.lc.filename = doc.lc.filename
        _errors__ = []
"""  # noqa: B907
        )

        self.idfield = idfield

        self.serializer.write(
            """
    def save(
        self, top: bool = False, base_url: str = "", relative_uris: bool = True
    ) -> dict[str, Any]:
        r: dict[str, Any] = {}

        if relative_uris:
            for ef in self.extension_fields:
                r[prefix_url(ef, self.loadingOptions.vocab)] = self.extension_fields[ef]
        else:
            for ef in self.extension_fields:
                r[ef] = self.extension_fields[ef]
"""
        )

    def end_class(self, classname: str, field_names: list[str]) -> None:
        """Signal that we are done with this class."""
        if self.current_class_is_abstract:
            return

        self.out.write(
            fmt(
                """
extension_fields: dict[str, Any] = {{}}
for k in _doc.keys():
    if k not in cls.attrs:
        if not k:
            _errors__.append(
                ValidationException("mapping with implicit null key")
            )
        elif ":" in k:
            ex = expand_url(
                k, "", loadingOptions, scoped_id=False, vocab_term=False
            )
            extension_fields[ex] = _doc[k]
        else:
            _errors__.append(
                ValidationException(
                    "invalid field `{{}}`, expected one of: {attrstr}".format(
                        k
                    ),
                    SourceLine(_doc, k, str),
                )
            )

if _errors__:
    raise ValidationException("", None, _errors__, "*")
""".format(
                    attrstr=", ".join([f"`{f}`" for f in field_names]),
                ),
                8,
            )
        )

        self.serializer.write(
            """
        # top refers to the directory level
        if top:
            if self.loadingOptions.namespaces:
                r["$namespaces"] = self.loadingOptions.namespaces
            if self.loadingOptions.schemas:
                r["$schemas"] = self.loadingOptions.schemas
"""
        )

        self.serializer.write("        return r\n\n")

        self.serializer.write(
            fmt(
                f"""attrs = frozenset(["{'", "'.join(field_names)}"])\n""",  # noqa: B907
                4,
            )
        )

        safe_init_fields: list[str] = [self.safe_name(f) for f in field_names if f != "class"]

        safe_inits = [f + "=" + f for f in safe_init_fields]

        safe_inits.extend(["extension_fields=extension_fields", "loadingOptions=loadingOptions"])

        self.out.write(
            "        _constructed = cls(\n            "
            + ",\n            ".join(safe_inits)
            + ",\n        )\n"
        )
        if self.idfield:
            self.out.write(
                f"        loadingOptions.idx[cast(str, {self.safe_name(self.idfield)})] "
                "= (_constructed, loadingOptions)\n"
            )

        self.out.write("        return _constructed\n")

        self.out.write(str(self.serializer.getvalue()))

        self.out.write("\n\n")

    def type_loader(
        self,
        type_declaration: Union[list[Any], dict[str, Any], str],
        container: Optional[str] = None,
        no_link_check: Optional[bool] = None,
    ) -> TypeDef:
        """Parse the given type declaration and declare its components."""
        sub_names: list[str]
        if isinstance(type_declaration, MutableSequence):
            sub_names = list(dict.fromkeys([self.type_loader(i).name for i in type_declaration]))
            return self.declare_type(
                TypeDef(
                    "union_of_{}".format("_or_".join(sub_names)),
                    "_UnionLoader(({},))".format(", ".join(sub_names)),
                )
            )
        if isinstance(type_declaration, MutableMapping):
            if type_declaration["type"] in (
                "array",
                "https://w3id.org/cwl/salad#array",
            ):
                i = self.type_loader(type_declaration["items"])
                return self.declare_type(
                    TypeDef(
                        f"array_of_{i.name}",
                        f"_ArrayLoader({i.name})",
                    )
                )
            if type_declaration["type"] in (
                "map",
                "https://w3id.org/cwl/salad#map",
            ):
                i = self.type_loader(type_declaration["values"])
                name = (
                    self.safe_name(type_declaration["name"]) if "name" in type_declaration else None
                )
                anon_type = self.declare_type(
                    TypeDef(
                        f"map_of_{i.name}",
                        "_MapLoader({}, {}, {}, {})".format(
                            i.name,
                            f"'{name}'",  # noqa: B907
                            f"'{container}'" if container is not None else None,  # noqa: B907
                            no_link_check,
                        ),
                    )
                )
                if "name" in type_declaration:
                    return self.declare_type(
                        TypeDef(self.safe_name(type_declaration["name"]) + "Loader", anon_type.name)
                    )
                else:
                    return anon_type
            if type_declaration["type"] in ("enum", "https://w3id.org/cwl/salad#enum"):
                for sym in type_declaration["symbols"]:
                    self.add_vocab(shortname(sym), sym)
                if "doc" in type_declaration:
                    doc = type_declaration["doc"]
                    if isinstance(doc, MutableSequence):
                        formated_doc = "\n".join(doc)
                    else:
                        formated_doc = doc.strip()
                    docstring = f'\n"""\n{formated_doc}\n"""'
                else:
                    docstring = ""
                return self.declare_type(
                    TypeDef(
                        self.safe_name(type_declaration["name"]) + "Loader",
                        '_EnumLoader(("{}",), "{}"){}'.format(
                            '", "'.join(
                                schema.avro_field_name(sym) for sym in type_declaration["symbols"]
                            ),
                            self.safe_name(type_declaration["name"]),
                            docstring,
                        ),
                    )
                )

            if type_declaration["type"] in (
                "record",
                "https://w3id.org/cwl/salad#record",
            ):
                return self.declare_type(
                    TypeDef(
                        self.safe_name(type_declaration["name"]) + "Loader",
                        "_RecordLoader({}, {}, {})".format(
                            self.safe_name(type_declaration["name"]),
                            f"'{container}'" if container is not None else None,  # noqa: B907
                            no_link_check,
                        ),
                        abstract=type_declaration.get("abstract", False),
                    )
                )

            if type_declaration["type"] in (
                "union",
                "https://w3id.org/cwl/salad#union",
            ):
                # Declare the named loader to handle recursive union definitions
                loader_name = self.safe_name(type_declaration["name"]) + "Loader"
                loader_type = TypeDef(loader_name, f"_UnionLoader((), '{loader_name}')")
                self.declare_type(loader_type)
                # Parse inner types
                sub_names = list(
                    dict.fromkeys([self.type_loader(i).name for i in type_declaration["names"]])
                )
                # Register lazy initialization for the loader
                self.add_lazy_init(
                    LazyInitDef(
                        loader_name,
                        "{}.add_loaders(({},))".format(loader_name, ", ".join(sub_names)),
                    )
                )
                return loader_type
            raise SchemaException("wft {}".format(type_declaration["type"]))

        if type_declaration in prims:
            return prims[type_declaration]

        if type_declaration in ("Expression", "https://w3id.org/cwl/cwl#Expression"):
            return self.declare_type(
                TypeDef(
                    self.safe_name(type_declaration) + "Loader",
                    "_ExpressionLoader(str)",
                )
            )
        return self.collected_types[self.safe_name(type_declaration) + "Loader"]

    def declare_id_field(
        self,
        name: str,
        fieldtype: TypeDef,
        doc: Optional[str],
        optional: bool,
    ) -> None:
        if self.current_class_is_abstract:
            return

        self.declare_field(name, fieldtype, doc, True, "")

        if optional:
            opt = """{safename} = "_:" + str(_uuid__.uuid4())""".format(
                safename=self.safe_name(name)
            )
        else:
            opt = """_errors__.append(ValidationException("missing {fieldname}"))""".format(
                fieldname=shortname(name)
            )

        self.out.write(
            """
        __original_{safename}_is_none = {safename} is None
        if {safename} is None:
            if docRoot is not None:
                {safename} = docRoot
            else:
                {opt}
        if not __original_{safename}_is_none:
            baseuri = cast(str, {safename})
""".format(
                safename=self.safe_name(name), opt=opt
            )
        )

    def declare_field(
        self,
        name: str,
        fieldtype: TypeDef,
        doc: Optional[str],
        optional: bool,
        subscope: Optional[str],
    ) -> None:
        if self.current_class_is_abstract:
            return

        if optional:
            self.out.write(f"""        {self.safe_name(name)} = None\n""")
            self.out.write(f"""        if "{shortname(name)}" in _doc:\n""")  # noqa: B907
            spc = "    "
        else:
            spc = ""

        if subscope:
            self.out.write(
                """
{spc}        subscope_baseuri = expand_url('{subscope}', baseuri, loadingOptions, True)
""".format(
                    subscope=subscope, spc=spc
                )
            )
            baseurivar = "subscope_baseuri"
        else:
            baseurivar = "baseuri"

        self.out.write(
            f"""{spc}        try:
"""
        )

        if not optional:
            self.out.write(
                """{spc}            if _doc.get("{fieldname}") is None:
{spc}                raise ValidationException("missing required field `{fieldname}`", None, [])

""".format(
                    fieldname=shortname(name),
                    spc=spc,
                )
            )

        self.out.write(
            """{spc}            {safename} = load_field(
{spc}                _doc.get("{fieldname}"),
{spc}                {fieldtype},
{spc}                {baseurivar},
{spc}                loadingOptions,
{spc}                lc=_doc.get("{fieldname}")
{spc}            )
""".format(
                safename=self.safe_name(name),
                fieldname=shortname(name),
                fieldtype=fieldtype.name,
                baseurivar=baseurivar,
                spc=spc,
            )
        )

        if shortname(name) == "class":
            self.out.write(
                """
{spc}            if {safename} not in (cls.__name__, loadingOptions.vocab.get(cls.__name__)):
{spc}               raise ValidationException(f"tried `{{cls.__name__}}` but")
{spc}        except ValidationException as e:
{spc}               raise e
""".format(
                    safename=self.safe_name(name),
                    spc=spc,
                )
            )

        else:
            self.out.write(
                """
{spc}        except ValidationException as e:
{spc}            error_message, to_print, verb_tensage = parse_errors(str(e))

{spc}            if str(e) == "missing required field `{fieldname}`":
{spc}                _errors__.append(
{spc}                    ValidationException(
{spc}                        str(e),
{spc}                        None
{spc}                    )
{spc}                )
{spc}            else:
{spc}                val = _doc.get("{fieldname}")
{spc}                if error_message != str(e):
{spc}                    val_type = convert_typing(extract_type(type(val)))
{spc}                    _errors__.append(
{spc}                        ValidationException(
{spc}                            \"the `{fieldname}` field is not valid because:\",
{spc}                            SourceLine(_doc, "{fieldname}", str),
{spc}                            [ValidationException(f"Value is a {{val_type}}, "
{spc}                                                 f"but valid {{to_print}} for this field "
{spc}                                                 f"{{verb_tensage}} {{error_message}}",
{spc}                                                 detailed_message=f"Value `{{val}}` is a {{val_type}}, "
{spc}                                                 f"but valid {{to_print}} for this field "
{spc}                                                 f"{{verb_tensage}} {{error_message}}")],
{spc}                        )
{spc}                    )
{spc}                else:
{spc}                    _errors__.append(
{spc}                        ValidationException(
{spc}                            "the `{fieldname}` field is not valid because:",
{spc}                            SourceLine(_doc, "{fieldname}", str),
{spc}                            [e],
{spc}                            detailed_message=f"the `{fieldname}` field with value `{{val}}` "
{spc}                            "is not valid because:\",
{spc}                        )
{spc}                    )
""".format(
                    fieldname=shortname(name),
                    spc=spc,
                )
            )

        if name == self.idfield or not self.idfield:
            baseurl = "base_url"
        else:
            baseurl = f"self.{self.safe_name(self.idfield)}"

        if shortname(name) == "class":
            self.serializer.write(
                fmt(
                    """
if self.{safename} is not None:
    uri = self.loadingOptions.vocab[self.{safename}]
    if p := self.loadingOptions.rvocab.get(uri[:-len(self.{safename})]):
        uri = f"{{p}}:{{self.{safename}}}"
    else:
        uri = self.{safename}
    u = save_relative_uri(uri, {baseurl}, {scoped_id}, {ref_scope}, relative_uris)
    r["{fieldname}"] = u
""".format(
                        safename=self.safe_name(name),
                        fieldname=shortname(name).strip(),
                        baseurl=baseurl,
                        scoped_id=fieldtype.scoped_id,
                        ref_scope=fieldtype.ref_scope,
                    ),
                    8,
                )
            )
        elif fieldtype.is_uri:
            self.serializer.write(
                fmt(
                    """
if self.{safename} is not None:
    u = save_relative_uri(self.{safename}, {baseurl}, {scoped_id}, {ref_scope}, relative_uris)
    r["{fieldname}"] = u
""".format(
                        safename=self.safe_name(name),
                        fieldname=shortname(name).strip(),
                        baseurl=baseurl,
                        scoped_id=fieldtype.scoped_id,
                        ref_scope=fieldtype.ref_scope,
                    ),
                    8,
                )
            )
        else:
            self.serializer.write(
                fmt(
                    """
if self.{safename} is not None:
    r["{fieldname}"] = save(
        self.{safename}, top=False, base_url={baseurl}, relative_uris=relative_uris
    )
""".format(
                        safename=self.safe_name(name),
                        fieldname=shortname(name),
                        baseurl=baseurl,
                    ),
                    8,
                )
            )

    def uri_loader(
        self,
        inner: TypeDef,
        scoped_id: bool,
        vocab_term: bool,
        ref_scope: Optional[int],
        no_link_check: Optional[bool] = None,
    ) -> TypeDef:
        """Construct the TypeDef for the given URI loader."""
        return self.declare_type(
            TypeDef(
                f"uri_{inner.name}_{scoped_id}_{vocab_term}_{ref_scope}_{no_link_check}",
                f"_URILoader({inner.name}, {scoped_id}, {vocab_term}, {ref_scope}, {no_link_check})",
                is_uri=True,
                scoped_id=scoped_id,
                ref_scope=ref_scope,
            )
        )

    def idmap_loader(
        self, field: str, inner: TypeDef, map_subject: str, map_predicate: Optional[str]
    ) -> TypeDef:
        """Construct the TypeDef for the given mapped ID loader."""
        return self.declare_type(
            TypeDef(
                f"idmap_{self.safe_name(field)}_{inner.name}",
                f"_IdMapLoader({inner.name}, '{map_subject}', '{map_predicate}')",  # noqa: B907
            )
        )

    def typedsl_loader(self, inner: TypeDef, ref_scope: Optional[int]) -> TypeDef:
        """Construct the TypeDef for the given DSL loader."""
        return self.declare_type(
            TypeDef(
                f"typedsl_{self.safe_name(inner.name)}_{ref_scope}",
                f"_TypeDSLLoader({self.safe_name(inner.name)}, {ref_scope}, "  # noqa: B907
                f"'{self.salad_version}')",  # noqa: B907
            )
        )

    def secondaryfilesdsl_loader(self, inner: TypeDef) -> TypeDef:
        """Construct the TypeDef for secondary files."""
        return self.declare_type(
            TypeDef(
                f"secondaryfilesdsl_{inner.name}",
                f"_UnionLoader((_SecondaryDSLLoader({inner.name}), {inner.name},))",
            )
        )

    def epilogue(self, root_loader: TypeDef) -> None:
        """Trigger to generate the epilouge code."""

        self.out.write("_vocab = {\n")
        for k in sorted(self.vocab.keys()):
            self.out.write(f'    "{k}": "{self.vocab[k]}",\n')  # noqa: B907
        self.out.write("}\n")

        self.out.write("_rvocab = {\n")
        for k in sorted(self.vocab.keys()):
            self.out.write(f'    "{self.vocab[k]}": "{k}",\n')  # noqa: B907
        self.out.write("}\n\n")

        for _, collected_type in self.collected_types.items():
            if not collected_type.abstract:
                self.out.write(fmt(f"{collected_type.name} = {collected_type.init}\n", 0))
        self.out.write("\n")

        if self.lazy_inits:
            for lazy_init in self.lazy_inits.values():
                self.out.write(fmt(f"{lazy_init.init}\n", 0))
            self.out.write("\n")

        self.out.write(
            """
def load_document(
    doc: Any,
    baseuri: Optional[str] = None,
    loadingOptions: Optional[LoadingOptions] = None,
) -> Any:
    if baseuri is None:
        baseuri = file_uri(os.getcwd()) + "/"
    if loadingOptions is None:
        loadingOptions = LoadingOptions()
    result, metadata = _document_load(
        %(name)s,
        doc,
        baseuri,
        loadingOptions,
    )
    return result


def load_document_with_metadata(
    doc: Any,
    baseuri: Optional[str] = None,
    loadingOptions: Optional[LoadingOptions] = None,
    addl_metadata_fields: Optional[MutableSequence[str]] = None,
) -> Any:
    if baseuri is None:
        baseuri = file_uri(os.getcwd()) + "/"
    if loadingOptions is None:
        loadingOptions = LoadingOptions(fileuri=baseuri)
    return _document_load(
        %(name)s,
        doc,
        baseuri,
        loadingOptions,
        addl_metadata_fields=addl_metadata_fields,
    )


def load_document_by_string(
    string: Any,
    uri: str,
    loadingOptions: Optional[LoadingOptions] = None,
) -> Any:
    yaml = yaml_no_ts()
    result = yaml.load(string)
    add_lc_filename(result, uri)

    if loadingOptions is None:
        loadingOptions = LoadingOptions(fileuri=uri)

    result, metadata = _document_load(
        %(name)s,
        result,
        uri,
        loadingOptions,
    )
    return result


def load_document_by_yaml(
    yaml: Any,
    uri: str,
    loadingOptions: Optional[LoadingOptions] = None,
) -> Any:
    """
            '"""'
            """
    Shortcut to load via a YAML object.
    yaml: must be from ruamel.yaml.main.YAML.load with preserve_quotes=True
    """
            '"""'
            """
    add_lc_filename(yaml, uri)

    if loadingOptions is None:
        loadingOptions = LoadingOptions(fileuri=uri)

    result, metadata = _document_load(
        %(name)s,
        yaml,
        uri,
        loadingOptions,
    )
    return result
"""
            % dict(name=root_loader.name)
        )



"""DotNet code generator for a given schema salad definition."""

import os
import shutil
import string
from collections.abc import MutableMapping, MutableSequence
from io import StringIO
from pathlib import Path
from typing import Any, Optional, Union
from xml.sax.saxutils import escape  # nosec

from . import _logger, schema
from .codegen_base import CodeGenBase, LazyInitDef, TypeDef
from .exceptions import SchemaException
from .java_codegen import _ensure_directory_and_write, _safe_makedirs
from .schema import shortname
from .utils import Traversable, files


def doc_to_doc_string(doc: Optional[str], indent_level: int = 0) -> str:
    """Generate a documentation string from a schema salad doc field."""
    lead = "" + "    " * indent_level + "/// "
    if doc:
        doc_str = "\n".join([f"{lead}{escape(line)}" for line in doc.split("\n")])
    else:
        doc_str = ""
    return doc_str


_string_type_def = TypeDef(
    instance_type="string",
    init="new PrimitiveLoader<string>()",
    name="StringInstance",
    loader_type="ILoader<string>",
)

_int_type_def = TypeDef(
    instance_type="int",
    init="new PrimitiveLoader<int>()",
    name="IntegerInstance",
    loader_type="ILoader<int>",
)

_long_type_def = TypeDef(
    instance_type="long",
    name="LongInstance",
    loader_type="ILoader<long>",
    init="new PrimitiveLoader<long>()",
)

_float_type_def = TypeDef(
    instance_type="double",
    name="DoubleInstance",
    loader_type="ILoader<double>",
    init="new PrimitiveLoader<double>()",
)

_bool_type_def = TypeDef(
    instance_type="bool",
    name="BooleanInstance",
    loader_type="ILoader<bool>",
    init="new PrimitiveLoader<bool>()",
)

_null_type_def = TypeDef(
    instance_type="None",
    name="NullInstance",
    loader_type="ILoader<object>",
    init="new NullLoader()",
)

_any_type_def = TypeDef(
    instance_type="object",
    name="AnyInstance",
    init="new AnyLoader()",
    loader_type="ILoader<object>",
)

prims = {
    "http://www.w3.org/2001/XMLSchema#string": _string_type_def,
    "http://www.w3.org/2001/XMLSchema#int": _int_type_def,
    "http://www.w3.org/2001/XMLSchema#long": _long_type_def,
    "http://www.w3.org/2001/XMLSchema#float": _float_type_def,
    "http://www.w3.org/2001/XMLSchema#double": _float_type_def,
    "http://www.w3.org/2001/XMLSchema#boolean": _bool_type_def,
    "https://w3id.org/cwl/salad#null": _null_type_def,
    "https://w3id.org/cwl/salad#Any": _any_type_def,
    "string": _string_type_def,
    "int": _int_type_def,
    "long": _long_type_def,
    "float": _float_type_def,
    "double": _float_type_def,
    "boolean": _bool_type_def,
    "null": _null_type_def,
    "Any": _any_type_def,
}


class DotNetCodeGen(CodeGenBase):
    """Generation of TypeScript code for a given Schema Salad definition."""

    def __init__(
        self, base: str, examples: Optional[str], target: Optional[str], package: str
    ) -> None:
        """Initialize the TypeScript codegen."""
        super().__init__()
        self.target_dir = Path(target or ".").resolve()
        self.main_src_dir = self.target_dir / package / "src"
        self.test_src_dir = self.target_dir / "Test"
        self.test_resources_dir = self.test_src_dir / "data"
        self.package = package
        self.base_uri = base
        self.examples = examples

    def prologue(self) -> None:
        """Trigger to generate the prolouge code."""
        for src_dir in [self.main_src_dir]:
            _safe_makedirs(src_dir)

        for primitive in prims.values():
            self.declare_type(primitive)

    @staticmethod
    def safe_name(name: str) -> str:
        """Generate a safe version of the given name."""
        avn = schema.avro_field_name(name)
        if avn.startswith("anon."):
            avn = avn[5:]
        if avn in (
            "class",
            "in",
            "extends",
            "abstract",
            "default",
            "package",
            "arguments",
            "out",
        ):
            # reserved words
            avn = avn + "_"

        return avn

    def begin_class(
        self,  # pylint: disable=too-many-arguments
        classname: str,
        extends: MutableSequence[str],
        doc: str,
        abstract: bool,
        field_names: MutableSequence[str],
        idfield: str,
        optional_fields: set[str],
    ) -> None:
        """Produce the header for the given class."""
        self.current_interface = "I" + self.safe_name(classname)
        cls = self.safe_name(classname)
        self.current_class = cls
        self.current_class_is_abstract = abstract
        interface_module_name = self.current_interface
        self.current_interface_target_file = self.main_src_dir / f"{interface_module_name}.cs"
        class_module_name = self.current_class
        self.current_class_target_file = self.main_src_dir / f"{class_module_name}.cs"
        self.current_constructor_signature = StringIO()
        self.current_constructor_signature_optionals = StringIO()
        self.current_constructor_body = StringIO()
        self.current_loader = StringIO()
        self.current_serializer = StringIO()
        self.current_fieldtypes: dict[str, TypeDef] = {}
        self.optional_field_names: list[str] = []
        self.mandatory_field_names: list[str] = []
        self.idfield = idfield

        doc_string = f"""
/// <summary>
/// Auto-generated interface for {classname}
"""
        if doc:
            doc_string += "///\n"
            doc_string += doc_to_doc_string(doc)
            doc_string += "\n"
        doc_string += "/// </summary>"

        with open(self.current_interface_target_file, "w") as f:
            _logger.info("Writing file: %s", self.current_interface_target_file)
            if extends:
                ext = " : " + ", ".join("I" + self.safe_name(e) for e in extends)
            else:
                ext = ""
            f.write(
                """#pragma warning disable CS0108
namespace {package};
{docstring}
public interface {cls}{ext}
{{
""".format(
                    docstring=doc_string,
                    cls=f"{self.current_interface}",
                    ext=ext,
                    package=self.package,
                )
            )
        if self.current_class_is_abstract:
            return

        doc_string = f"""
/// <summary>
/// Auto-generated class implementation for {classname}
"""
        if doc:
            doc_string += "///\n"
            doc_string += doc_to_doc_string(doc)
            doc_string += "\n"
        doc_string += "/// </summary>"
        with open(self.current_class_target_file, "w") as f:
            _logger.info("Writing file: %s", self.current_class_target_file)
            f.write(
                """using System.Collections;
using OneOf;
using OneOf.Types;

namespace {package};
{docstring}
public class {cls} : {current_interface}, ISaveable
{{
    readonly LoadingOptions loadingOptions;

    readonly Dictionary<object, object> extensionFields;
""".format(
                    cls=cls,
                    current_interface=self.current_interface,
                    docstring=doc_string,
                    package=self.package,
                )
            )
        self.current_constructor_signature.write(
            "\n"
            + "\n"
            + "    public {cls}(".format(
                cls=cls,
            )
        )
        self.current_constructor_body.write(
            """
        this.loadingOptions = loadingOptions ?? new LoadingOptions();
        this.extensionFields = extensionFields ?? new Dictionary<object, object>();
"""
        )
        self.current_loader.write(
            """
    public static ISaveable FromDoc(object doc__, string baseUri, LoadingOptions loadingOptions,
        string? docRoot = null)
    {
        List<ValidationException> errors = new();

        if (doc__ is not IDictionary)
        {
            throw new ValidationException("Document has to be of type Dictionary");
        }

        Dictionary<object, object> doc_ = ((IDictionary)doc__)
            .Cast<dynamic>()
            .ToDictionary(entry => entry.Key, entry => entry.Value);
"""
        )

        self.current_serializer.write(
            """
    public Dictionary<object, object> Save(bool top = false, string baseUrl = "",
        bool relativeUris = true)
    {
        Dictionary<object, object> r = new();
        foreach (KeyValuePair<object, object> ef in extensionFields)
        {
            r[loadingOptions.PrefixUrl((string)ef.Value)] = ef.Value;
        }
"""
        )

    def end_class(self, classname: str, field_names: list[str]) -> None:
        """Signal that we are done with this class."""
        with open(self.current_interface_target_file, "a") as f:
            f.write("}\n")
        if self.current_class_is_abstract:
            return

        self.current_constructor_signature.write(
            self.current_constructor_signature_optionals.getvalue()
        )
        self.current_constructor_signature.write(
            "LoadingOptions? loadingOptions = null, "
            "Dictionary<object, object>? extensionFields = null)"
            "\n    "
            "{"
        )
        self.current_constructor_body.write("    }\n")
        self.current_loader.write(
            """
        Dictionary<object, object> extensionFields = new();
        foreach (KeyValuePair<object, object> v in doc_)
        {{
            if (!attr.Contains(v.Key))
            {{
                if (((string)v.Key).Contains(':'))
                {{
                    string ex = loadingOptions.ExpandUrl((string)v.Key, "", false, false, null);
                    extensionFields[ex] = v.Value;
                }}
                else
                {{
                    errors.Add(
                        new ValidationException($"invalid field {{v.Key}}," +
                        "expected one of {fields}"));
                    break;
                }}
            }}

        }}

        if (errors.Count > 0)
        {{
            throw new ValidationException("", errors);
        }}

        {classname} res__ = new(
          """.format(
                classname=self.current_class,
                fields=", ".join(["`" + f + "`" for f in field_names]),
            )
        )
        self.current_loader.write("loadingOptions: loadingOptions")
        if len(self.mandatory_field_names) > 0:
            self.current_loader.write(
                ",\n          "
                + ",\n          ".join(f + ": " + f for f in self.mandatory_field_names)
            )
        self.current_loader.write("\n        );\n")

        for optionalField in self.optional_field_names:
            self.current_loader.write(
                f"""
        if ({optionalField} != null)
        {{
            res__.{optionalField} = {optionalField};
        }}
"""
            )
        self.current_loader.write("\n        return res__;")
        self.current_loader.write("\n    " + "}" + "\n")
        self.current_serializer.write(
            """
        if (top)
        {
            if (loadingOptions.namespaces != null)
            {
                r["$namespaces"] = loadingOptions.namespaces;
            }

            if (this.loadingOptions.schemas != null)
            {
                r["$schemas"] = loadingOptions.schemas;
            }
        }

        return r;
    }

"""
        )
        with open(
            self.current_class_target_file,
            "a",
        ) as f:
            f.write(self.current_constructor_signature.getvalue())
            f.write(self.current_constructor_body.getvalue())
            f.write(self.current_loader.getvalue())
            f.write(self.current_serializer.getvalue())
            f.write(
                "    static readonly System.Collections.Generic.HashSet<string>"
                + " attr = new() { "
                + ", ".join(['"' + shortname(f) + '"' for f in field_names])
                + " };"
            )
            f.write(
                """
}
"""
            )

    def type_loader(
        self,
        type_declaration: Union[list[Any], dict[str, Any], str],
        container: Optional[str] = None,
        no_link_check: Optional[bool] = None,
    ) -> TypeDef:
        """Parse the given type declaration and declare its components."""
        if isinstance(type_declaration, MutableSequence):
            sub_types = [self.type_loader(i) for i in type_declaration]
            sub_names: list[str] = list(dict.fromkeys([i.name for i in sub_types]))
            sub_instance_types: list[str] = list(
                dict.fromkeys([i.instance_type for i in sub_types if i.instance_type is not None])
            )
            return self.declare_type(
                TypeDef(
                    name="union_of_{}".format("_or_".join(sub_names)),
                    init="new UnionLoader(new List<ILoader> {{ {} }})".format(", ".join(sub_names)),
                    instance_type="OneOf<" + ", ".join(sub_instance_types) + ">",
                    loader_type="ILoader<object>",
                )
            )
        if isinstance(type_declaration, MutableMapping):
            if type_declaration["type"] in (
                "array",
                "https://w3id.org/cwl/salad#array",
            ):
                i = self.type_loader(type_declaration["items"])
                return self.declare_type(
                    TypeDef(
                        instance_type=f"List<{i.instance_type}>",
                        name=f"array_of_{i.name}",
                        loader_type=f"ILoader<List<{i.instance_type}>>",
                        init=f"new ArrayLoader<{i.instance_type}>({i.name})",
                    )
                )
            if type_declaration["type"] in (
                "map",
                "https://w3id.org/cwl/salad#map",
            ):
                i = self.type_loader(type_declaration["values"])
                return self.declare_type(
                    TypeDef(
                        instance_type=f"Dictionary<string, {i.instance_type}>",
                        name=f"map_of_{i.name}",
                        loader_type=f"ILoader<Dictionary<string, {i.instance_type}>>",
                        init="new MapLoader<{}>({}, {}, {})".format(
                            i.instance_type,
                            i.name,
                            (
                                f"'{container}'" if container is not None else self.to_dotnet(None)
                            ),  # noqa: B907
                            self.to_dotnet(no_link_check),
                        ),
                    )
                )
            if type_declaration["type"] in ("enum", "https://w3id.org/cwl/salad#enum"):
                return self.type_loader_enum(type_declaration)
            if type_declaration["type"] in (
                "record",
                "https://w3id.org/cwl/salad#record",
            ):
                return self.declare_type(
                    TypeDef(
                        instance_type=self.safe_name(type_declaration["name"]),
                        name=self.safe_name(type_declaration["name"]) + "Loader",
                        init="new RecordLoader<{}>({}, {})".format(
                            self.safe_name(type_declaration["name"]),
                            (
                                f"'{container}'" if container is not None else self.to_dotnet(None)
                            ),  # noqa: B907
                            self.to_dotnet(no_link_check),
                        ),
                        loader_type="ILoader<{}>".format(self.safe_name(type_declaration["name"])),
                        abstract=type_declaration.get("abstract", False),
                    )
                )
            if type_declaration["type"] in (
                "union",
                "https://w3id.org/cwl/salad#union",
            ):
                # Declare the named loader to handle recursive union definitions
                loader_name = self.safe_name(type_declaration["name"]) + "Loader"
                loader_type = TypeDef(
                    name=loader_name,
                    init="new UnionLoader(new List<ILoader>())",
                    instance_type="object",
                    loader_type="ILoader<object>",
                )
                self.declare_type(loader_type)
                # Parse inner types
                sub_types = [self.type_loader(i) for i in type_declaration["names"]]
                # Register lazy initialization for the loader
                self.add_lazy_init(
                    LazyInitDef(
                        loader_name,
                        "((UnionLoader){}).addLoaders(new List<ILoader> {{ {} }});".format(
                            loader_name, ", ".join(s.name for s in sub_types)
                        ),
                    )
                )
                return loader_type
            raise SchemaException("wft {}".format(type_declaration["type"]))
        if type_declaration in prims:
            return prims[type_declaration]
        if type_declaration in ("Expression", "https://w3id.org/cwl/cwl#Expression"):
            return self.declare_type(
                TypeDef(
                    name=self.safe_name(type_declaration) + "Loader",
                    init="new ExpressionLoader()",
                    loader_type="ILoader<string>",
                    instance_type="string",
                )
            )
        return self.collected_types[self.safe_name(type_declaration) + "Loader"]

    def type_loader_enum(self, type_declaration: dict[str, Any]) -> TypeDef:
        """Build an enum type loader for the given declaration."""
        for sym in type_declaration["symbols"]:
            self.add_vocab(shortname(sym), sym)
        enum_name = self.safe_name(type_declaration["name"])
        enum_module_name = enum_name
        enum_path = self.main_src_dir / f"{enum_module_name}.cs"
        with open(enum_path, "w") as f:
            _logger.info("Writing file: %s", enum_path)
            f.write(
                """namespace {package};

public class {enum_name} : IEnumClass<{enum_name}>
{{
    private string _Name;
    private static readonly List<{enum_name}> members = new();

""".format(
                    enum_name=enum_name, package=self.package
                )
            )
            for sym in type_declaration["symbols"]:
                const = self.safe_name(sym).replace("-", "_").replace(".", "_").upper()
                f.write(
                    """    public static readonly {enum_name} {const} =
                            new("{val}");\n""".format(
                        const=const, val=self.safe_name(sym), enum_name=enum_name
                    )
                )
            f.write(
                """
    public string Name
    {{
        get {{ return _Name; }}
        private set {{ _Name = value; }}
    }}

    public static IList<{enum_name}> Members
    {{
        get {{ return members; }}
    }}

    private {enum_name}(string name)
    {{
        _Name = name;
        members.Add(this);
    }}

    public static {enum_name} Parse(string toParse)
    {{
        foreach ({enum_name} s in Members)
        {{
            if (toParse == s.Name)
                return s;
        }}

        throw new FormatException("Could not parse string.");
    }}

    public static bool Contains(string value)
    {{
        bool contains = false;
        foreach ({enum_name} s in Members)
        {{
            if (value == s.Name)
            {{
                contains = true;
                return contains;
            }}

        }}

        return contains;
    }}

    public static List<string> Symbols()
    {{
        return members.Select(m => m.Name).ToList();
    }}

    public override string ToString()
    {{
        return _Name;
    }}
}}
""".format(
                    enum_name=enum_name
                )
            )
        return self.declare_type(
            TypeDef(
                instance_type=enum_name,
                name=self.safe_name(type_declaration["name"] + "Loader"),
                init=f"new EnumLoader<{enum_name}>()",
                loader_type=f"ILoader<{enum_name}>",
            )
        )

    def declare_field(
        self,
        name: str,
        fieldtype: TypeDef,
        doc: Optional[str],
        optional: bool,
        subscope: Optional[str],
    ) -> None:
        """Output the code to load the given field."""
        if self.current_class_is_abstract:
            return
        safename = self.safe_name(name)
        fieldname = shortname(name)
        self.current_fieldtypes[safename] = fieldtype

        if optional:
            self.optional_field_names.append(safename)
            if fieldtype.instance_type is not None and not fieldtype.instance_type.startswith(
                "OneOf<None"
            ):
                optionalstring = "?"
            else:
                optionalstring = ""
        else:
            self.mandatory_field_names.append(safename)
            optionalstring = ""

        with open(self.current_class_target_file, "a") as f:
            if doc:
                f.write(
                    """
    /// <summary>
{doc_str}
    /// </summary>
""".format(
                        doc_str=doc_to_doc_string(doc, indent_level=1)
                    )
                )
            f.write(
                "    public {type}{optionalstring} {safename} {{ get; set; }}\n".format(
                    safename=safename,
                    type=fieldtype.instance_type,
                    optionalstring=optionalstring,
                )
            )
        if fieldname == "class":
            if fieldtype.instance_type == "string":
                self.current_constructor_signature_optionals.write(
                    'string {safename} = "{val}", '.format(
                        safename=safename, val=self.current_class
                    )
                )
            else:
                self.current_constructor_signature_optionals.write(
                    "{type}? {safename} = null, ".format(
                        safename=safename, type=fieldtype.instance_type
                    )
                )
        else:
            if not optional:
                self.current_constructor_signature.write(
                    "{type} {safename}, ".format(
                        safename=safename,
                        type=fieldtype.instance_type,
                    )
                )
            else:
                if fieldtype.instance_type is not None and fieldtype.instance_type.startswith(
                    "OneOf<None"
                ):
                    self.current_constructor_signature_optionals.write(
                        "{type} {safename} = default, ".format(
                            safename=safename,
                            type=fieldtype.instance_type,
                        )
                    )
                else:
                    self.current_constructor_signature_optionals.write(
                        "{type}? {safename} = null, ".format(
                            safename=safename,
                            type=fieldtype.instance_type,
                        )
                    )
        if fieldname == "class" and fieldtype.instance_type != "string":
            self.current_constructor_body.write(
                "        this.{safeName} = {safeName} ?? {type}.{val};\n".format(
                    safeName=safename,
                    type=fieldtype.instance_type,
                    val=self.current_class.replace("-", "_").replace(".", "_").upper(),
                )
            )

        else:
            self.current_constructor_body.write(
                "        this.{safeName} = {safeName};\n".format(safeName=safename)
            )

        self.current_loader.write(
            """
        dynamic {safename} = default!;""".format(
                safename=safename
            )
        )
        if optional:
            self.current_loader.write(
                """
        if (doc_.ContainsKey("{fieldname}"))
        {{""".format(
                    fieldname=fieldname
                )
            )
            spc = "        "
        else:
            spc = "    "

        self.current_loader.write(
            """
{spc}    try
{spc}    {{
{spc}        {safename} = LoaderInstances.{fieldtype}
{spc}           .LoadField(doc_.GetValueOrDefault("{fieldname}", null!), baseUri,
{spc}               loadingOptions);
{spc}    }}
{spc}    catch (ValidationException e)
{spc}    {{
{spc}        errors.Add(
{spc}          new ValidationException("the `{fieldname}` field is not valid because: ", e)
{spc}        );
{spc}    }}
""".format(
                safename=safename,
                fieldname=fieldname,
                fieldtype=fieldtype.name,
                spc=spc,
            )
        )
        if optional:
            self.current_loader.write("        }\n")

        if name == self.idfield or not self.idfield:
            baseurl = "baseUrl"
        elif self.id_field_type.instance_type is not None:
            if self.id_field_type.instance_type.startswith("OneOf"):
                baseurl = (
                    f"(this.{self.safe_name(self.idfield)}.Value is "
                    f'None ? "" : {self.safe_name(self.idfield)}.Value)'
                )
            else:
                baseurl = f"this.{self.safe_name(self.idfield)}"

        if fieldtype.is_uri:
            self.current_serializer.write(
                """
        object? {safename}Val = ISaveable.SaveRelativeUri({safename}, {scoped_id},
            relativeUris, {ref_scope}, (string){base_url}!);
        if ({safename}Val is not null)
        {{
            r["{fieldname}"] = {safename}Val;
        }}
""".format(
                    safename=self.safe_name(name),
                    fieldname=shortname(name).strip(),
                    base_url=baseurl,
                    scoped_id=self.to_dotnet(fieldtype.scoped_id),
                    ref_scope=self.to_dotnet(fieldtype.ref_scope),
                )
            )
        else:
            self.current_serializer.write(
                """
        object? {safename}Val = ISaveable.Save({safename},
                                        false, (string){base_url}!, relativeUris);
        if ({safename}Val is not null)
        {{
            r["{fieldname}"] = {safename}Val;
        }}
""".format(
                    safename=self.safe_name(name),
                    fieldname=shortname(name).strip(),
                    base_url=baseurl,
                )
            )

    def declare_id_field(
        self,
        name: str,
        fieldtype: TypeDef,
        doc: Optional[str],
        optional: bool,
    ) -> None:
        """Output the code to handle the given ID field."""
        self.id_field_type = fieldtype
        if self.current_class_is_abstract:
            return
        self.declare_field(name, fieldtype, doc, True, "")
        if optional:
            opt = f"""{self.safe_name(name)} = "_" + Guid.NewGuid();"""
        else:
            opt = """throw new ValidationException("Missing {fieldname}");""".format(
                fieldname=shortname(name)
            )

        self.current_loader.write(
            """
        if ({safename} == null)
        {{
            if (docRoot != null)
            {{
                {safename} = docRoot;
            }}
            else
            {{
                {opt}
            }}
        }}
        else
        {{
            baseUri = (string){safename};
        }}
""".format(
                safename=self.safe_name(name), opt=opt
            )
        )

    def to_dotnet(self, val: Any) -> Any:
        """Convert a Python keyword to a DotNet keyword."""
        if val is True:
            return "true"
        elif val is None:
            return "null"
        elif val is False:
            return "false"
        return val

    def uri_loader(
        self,
        inner: TypeDef,
        scoped_id: bool,
        vocab_term: bool,
        ref_scope: Optional[int],
        no_link_check: Optional[bool] = None,
    ) -> TypeDef:
        """Construct the TypeDef for the given URI loader."""
        instance_type = inner.instance_type or "object"
        return self.declare_type(
            TypeDef(
                instance_type=instance_type,
                name=f"uri{inner.name}{scoped_id}{vocab_term}{ref_scope}{no_link_check}",
                loader_type="ILoader<object>",
                init="new UriLoader({}, {}, {}, {}, {})".format(
                    inner.name,
                    self.to_dotnet(scoped_id),
                    self.to_dotnet(vocab_term),
                    self.to_dotnet(ref_scope),
                    self.to_dotnet(no_link_check),
                ),
                is_uri=True,
                scoped_id=scoped_id,
                ref_scope=ref_scope,
            )
        )

    def idmap_loader(
        self, field: str, inner: TypeDef, map_subject: str, map_predicate: Optional[str]
    ) -> TypeDef:
        """Construct the TypeDef for the given mapped ID loader."""
        instance_type = inner.instance_type or "object"
        return self.declare_type(
            TypeDef(
                instance_type=instance_type,
                name=f"idmap{self.safe_name(field)}{inner.name}",
                loader_type="ILoader<object>",
                init='new IdMapLoader({}, "{}", "{}")'.format(
                    inner.name, map_subject, map_predicate
                ),
            )
        )

    def typedsl_loader(self, inner: TypeDef, ref_scope: Optional[int]) -> TypeDef:
        """Construct the TypeDef for the given DSL loader."""
        instance_type = inner.instance_type or "object"
        return self.declare_type(
            TypeDef(
                instance_type=instance_type,
                name=f"typedsl{self.safe_name(inner.name)}{ref_scope}",
                loader_type="ILoader<object>",
                init=(f"new TypeDSLLoader" f"({self.safe_name(inner.name)}, {ref_scope})"),
            )
        )

    def epilogue(self, root_loader: TypeDef) -> None:
        """Trigger to generate the epilouge code."""
        pd = "This project contains .Net objects and utilities "
        pd = pd + ' auto-generated by <a href="https://github.com/'
        pd = pd + 'common-workflow-language/schema_salad">Schema Salad</a>'
        pd = pd + " for parsing documents corresponding to the "
        pd = pd + str(self.base_uri) + " schema."

        template_vars: MutableMapping[str, str] = dict(
            project_name=self.package,
            version="0.0.1-SNAPSHOT",
            project_description=pd,
            license_name="Apache License, Version 2.0",
        )

        def template_from_resource(resource: Traversable) -> string.Template:
            template_str = resource.read_text("utf-8")
            template = string.Template(template_str)
            return template

        def expand_resource_template_to(resource: str, path: Path) -> None:
            template = template_from_resource(files("schema_salad").joinpath(f"dotnet/{resource}"))
            src = template.safe_substitute(template_vars)
            _ensure_directory_and_write(path, src)

        expand_resource_template_to("editorconfig", self.target_dir / ".editorconfig")
        expand_resource_template_to("gitignore", self.target_dir / ".gitignore")
        expand_resource_template_to("LICENSE", self.target_dir / "LICENSE")
        expand_resource_template_to("README.md", self.target_dir / "README.md")
        expand_resource_template_to("Solution.sln", self.target_dir / "Solution.sln")
        expand_resource_template_to(
            "Project.csproj.template",
            self.target_dir / self.package / Path(self.package + ".csproj"),
        )
        expand_resource_template_to(
            "Test.csproj.template",
            self.test_src_dir / "Test.csproj",
        )
        expand_resource_template_to(
            "docfx.json",
            self.target_dir / self.package / "docfx.json",
        )
        expand_resource_template_to(
            "AssemblyInfo.cs",
            self.target_dir / self.package / "Properties" / "AssemblyInfo.cs",
        )
        vocab = ",\n        ".join(
            f"""["{k}"] = "{self.vocab[k]}\"""" for k in sorted(self.vocab.keys())  # noqa: B907
        )
        rvocab = ",\n        ".join(
            f"""["{self.vocab[k]}"] = "{k}\"""" for k in sorted(self.vocab.keys())  # noqa: B907
        )

        loader_instances = ""
        for _, collected_type in self.collected_types.items():
            if not collected_type.abstract:
                loader_instances += "    internal static readonly {} {} = {};\n".format(
                    collected_type.loader_type, collected_type.name, collected_type.init
                )

        if self.lazy_inits:
            loader_instances += "\n    static LoaderInstances()\n    {\n"
            for lazy_init in self.lazy_inits.values():
                loader_instances += f"        {lazy_init.init}\n"
            loader_instances += "    }\n"

        example_tests = ""
        if self.examples:
            _safe_makedirs(self.test_resources_dir)
            utils_resources = self.test_resources_dir / "examples"
            if os.path.exists(utils_resources):
                shutil.rmtree(utils_resources)
            shutil.copytree(self.examples, utils_resources)
            for example_name in os.listdir(self.examples):
                if example_name.startswith("valid"):
                    basename = os.path.basename(example_name).rsplit(".", 1)[0]
                    example_tests += """
    [TestMethod]
    public void Test{basename}()
    {{
        string? file = System.IO.File.ReadAllText("data/examples/{example_name}");
        RootLoader.LoadDocument(file!,
            new Uri(Path.GetFullPath("data/examples/{example_name}")).AbsoluteUri);
    }}
""".format(
                        basename=basename.replace("-", "_").replace(".", "_"),
                        example_name=example_name,
                    )

        template_args: MutableMapping[str, str] = dict(
            project_name=self.package,
            loader_instances=loader_instances,
            vocab=vocab,
            rvocab=rvocab,
            root_loader=root_loader.name,
            root_loader_type=root_loader.instance_type or "object",
            tests=example_tests,
            project_description=pd,
        )

        util_src_dirs = {
            "util": self.main_src_dir / "util",
            "Test": self.test_src_dir,
            "DocFx": self.target_dir / "DocFx",
        }

        def copy_utils_recursive(util_src: str, util_target: Path) -> None:
            for util in files("schema_salad").joinpath(f"dotnet/{util_src}").iterdir():
                if util.is_dir():
                    copy_utils_recursive(os.path.join(util_src, util.name), util_target / util.name)
                    continue
                src_path = util_target / util.name
                src_template = template_from_resource(util)
                src = src_template.safe_substitute(template_args)
                _ensure_directory_and_write(src_path, src)

        for util_src, util_target in util_src_dirs.items():
            copy_utils_recursive(util_src, util_target)

    def secondaryfilesdsl_loader(self, inner: TypeDef) -> TypeDef:
        """Construct the TypeDef for secondary files."""
        instance_type = inner.instance_type or "any"
        return self.declare_type(
            TypeDef(
                name=f"secondaryfilesdsl{inner.name}",
                init=f"new SecondaryDSLLoader({inner.name})",
                loader_type="ILoader<object>",
                instance_type=instance_type,
            )
        )



"""Shared test functions and attributes."""

import atexit
import os
from contextlib import ExitStack
from pathlib import Path

from schema_salad.utils import as_file, files


def get_path(filename: str) -> Path:
    """Get the file path for a given schema file name.

    It is able to find file names in the ``schema_salad`` namespace, but
    also able to load schema files from the ``tests`` directory.
    """
    filename = os.path.normpath(filename)  # normalizing path depending on OS
    # or else it will cause problem when joining path
    filepath = None
    try:
        file_manager = ExitStack()
        atexit.register(file_manager.close)
        traversable = files("schema-salad") / filename
        filepath = file_manager.enter_context(as_file(traversable))
    except ModuleNotFoundError:
        pass
    if not filepath or not filepath.is_file():
        # First try to load it from the local directory, probably ``./tests/``.
        filepath = Path(os.path.dirname(__file__)) / filename
        if not filepath.is_file():
            # If that didn't work, then default to tests/../${filename},
            # note that we return the parent as it is expected that __file__
            # is a test file.
            filepath = Path(os.path.dirname(__file__)) / ".." / filename
    return filepath.resolve()


def get_data(filename: str) -> str:
    """Get the file path for a given schema file name.

    It is able to find file names in the ``schema_salad`` namespace, but
    also able to load schema files from the ``tests`` directory.
    """
    return str(get_path(filename))


def get_data_uri(resource_path: str) -> str:
    """Get the file URI for tests."""
    return get_path(resource_path).as_uri()


# Schemas used in tests

cwl_file_uri = get_data_uri("tests/test_schema/CommonWorkflowLanguage.yml")
metaschema_file_uri = get_data_uri("metaschema/metaschema.yml")
basket_file_uri = get_data_uri("tests/basket_schema.yml")
